using System.Reflection.Metadata;
using System.Linq;
using WingServerCommon.Service;
using WingInterfaceLibrary.Interface;
using System.Threading.Tasks;
using System.Collections.Generic;
using WingInterfaceLibrary.Request.Education;
using WingInterfaceLibrary.DB.Request;
using WingInterfaceLibrary.Enum;
using WingInterfaceLibrary.Enum.CourseEnum;
using WingInterfaceLibrary.Result.Education;
using WingInterfaceLibrary.Rtc;
using WingInterfaceLibrary.Internal.Request;
using WingServerCommon.Log;
using WingInterfaceLibrary.Notifications.Consultation;
using WingInterfaceLibrary.DTO.RTC;
using System;
using WingInterfaceLibrary.Request.DBRequest;
using WingServerCommon.Config;
using WingInterfaceLibrary.DTO.Education;
using WingServerCommon.Interfaces.Cache;
using WingInterfaceLibrary.DTO.LiveRoom;
using Newtonsoft.Json;
using WingInterfaceLibrary.LiveConsultation;
using WingServerCommon.Config.Parameters;
using WingInterfaceLibrary.LiveConsultation.Consultation;
using WingInterfaceLibrary.DTO.Courses;
using System.Timers;
using WingInterfaceLibrary.Request.Courses;
using System.Collections.Concurrent;
using WingServerCommon.Mapper;

namespace WingEducationService.Service
{
    /// <summary>
    /// 课程直播服务
    /// </summary>
    public partial class EducationService : JsonRpcService, IEducationService
    {
        /// <summary>
        /// 老师开始直播
        /// </summary>
        /// <param name="request">开始直播请求实体</param>
        /// <returns></returns>
        public async Task<InitiateLiveCourseResult> InitiateLiveCourseAsync(InitiateLiveCourseRequest request)
        {
            var teacherCode = await GetClientIdByTokenAsync(request.Token);
            //保存发起人
            var roomId = request.CourseCode;
            if (string.IsNullOrEmpty(roomId))
            {
                ThrowRpcException(CustomerRpcCode.CourseCodeEmpty, "Course code empty");
            };
            using (await _lockManager.Acquire(roomId))
            {
                // 发起人只能是老师
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                if (courseDetailInfo == null || courseDetailInfo.TeacherCode != teacherCode)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    Logger.WriteLineWarn($"WingEducationService initiate live course failed, course not existed, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
                };
                var courseName = courseDetailInfo.Name;
                if (courseDetailInfo.Status == CourseStatusEnum.NoApproval || courseDetailInfo.Status == CourseStatusEnum.RejectApproval)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] reservation record not reviewed, current status:{courseDetailInfo.Status}, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseNotReviewed, "Course not reviewed");
                }
                if (courseDetailInfo.Status == CourseStatusEnum.Started)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] course in process, current status:{courseDetailInfo.Status}, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseInProcess, "Course In Process");
                }
                if (courseDetailInfo.Status == CourseStatusEnum.HasEnded || courseDetailInfo.Status == CourseStatusEnum.Offline)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] course ended, current status:{courseDetailInfo.Status}, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseEnded, "Course ended");
                }
                //状态验证，老师/助教，一级学生是否在线, 建议优化，你本身获取房间的信息的时候，就应该所有数据都是新的
                var userCodes = new List<string>() { teacherCode };
                if (courseDetailInfo.Experts?.Count > 0)
                {
                    var expertUsers = courseDetailInfo.Experts.Select(c => c.Code).Distinct().ToList();
                    userCodes.AddRange(expertUsers);
                }
                if (courseDetailInfo.Assistants?.Count > 0)
                {
                    var assistantsUsers = courseDetailInfo.Assistants.Select(c => c.Code).Distinct().ToList();
                    userCodes.AddRange(assistantsUsers);
                }
                var memberRequest = new GetLiveMemberStatusInfosRequest()
                {
                    UserCodes = userCodes
                };
                var memberStautsInfo = await _rtcService.GetLiveMemberStatusInfosAsync(memberRequest);
                var roomInitiator = new LiveMemberStatusDTO();
                if (memberStautsInfo?.Count > 0)
                {
                    roomInitiator = memberStautsInfo.FirstOrDefault(c => c.Code == teacherCode) ?? new LiveMemberStatusDTO();
                }
                if (roomInitiator == null || roomInitiator.UserStatusType == UserStatusEnum.IsBusy)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] initiate live course failed, teacher not online, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.TeacherNotOnline, "Course Teacher not online");
                }
                if (memberStautsInfo.Count(x => x.UserStatusType == UserStatusEnum.Idle) < 1)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] initiate live course failed, other members not online, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.StudentsNotOnline, "Course students not online");
                }

                //修改预约记录状态
                var courseStatusRequest = new UpdateCourseStatusDBRequest()
                {
                    Code = roomId,
                    Status = CourseStatusEnum.Started
                };
                Logger.WriteLineWarn($"EducationService InitiateLiveCourseAsync Started 2, roomId:{roomId}");
                var updateResult = await _courseDBService.UpdateCourseStatusAsync(courseStatusRequest);
                if (!updateResult)
                {
                    Logger.WriteLineWarn($"WingEducationService [{room.Name}] update Status to db failed, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseDBError, "Course DB error");
                }

                var initiateLive = new InitiateRequest()
                {
                    LiveRoomCode = roomId,
                    InitiatorCode = teacherCode
                };
                room = await _rtcService.InitiateAsync(initiateLive);
                var userSignRequest = new GetUserSignRequest
                {
                    UserId = teacherCode
                };
                var userSignResult = await _wingRtcService.GetUserSignAsync(userSignRequest);
                var result = new InitiateLiveCourseResult
                {
                    CourseCode = roomId,
                    RoomNo = room.RtcRoomId,
                    LiveProtocol = EnvironmentConfigs.General.LiveProtocolDetail<TransactionStatusEnum>(),
                    InitiatorCode = teacherCode,
                    AppId = room.SdkAppId,
                    UserSign = userSignResult.UserSign,
                    HeartRateInterval = EnvironmentConfigManager.GetParammeter<IntParameter>("LiveConsultation", "HeartRateSeconds").Value,
                    MemberLiveDatas = TransferLiveMemberToCourseMember(room.UserInfos, teacherCode, memberStautsInfo, courseDetailInfo.Experts, courseDetailInfo.Assistants),
                    CourseChannel = courseDetailInfo.CourseChannel,
                    OthersScreenSharingChannel = courseDetailInfo.OthersScreenSharingChannel,
                };
                if (room.RtcRoomId > 0)
                {
                    var rtcScreenSharingChannel = await GenerateRtcInfoAsync(room.RtcRoomId, room.SdkAppId);
                    result.RtcScreenSharingChannel = rtcScreenSharingChannel;
                    Logger.WriteLineWarn($"EducationService InitiateLiveCourseAsync Started, roomId:{roomId}");
                    var updateCourseResult = await _courseDBService.UpdateCourseStatusAsync(new UpdateCourseStatusDBRequest()
                    {
                        Code = roomId,
                        Status = CourseStatusEnum.Started,
                        RtcScreenSharingChannel = rtcScreenSharingChannel,
                    });
                }

                return result;
            }
        }

        /// <summary>
        /// Cancel the initiating a diagnosis meeting
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        /// <show>false</show>
        public async Task<LiveCourseBaseResult> CancelLiveCourseAsync(WingInterfaceLibrary.Request.Education.CancelLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            using (await _lockManager.Acquire(roomId))
            {
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                var room = await _rtcService.GetLiveRoomAsync(new GetLiveRoomRequest { LiveRoomCode = roomId });
                if (room == null || courseDetailInfo == null)
                {
                    ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
                }
                if (room.Status == LiveRoomStatus.Cancelled)
                {
                    ThrowRpcException(CustomerRpcCode.LiveCourseCancelled, "Live course cancelled");
                }
                if (userCode != room.InitiatorCode)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                //修改课程记录状态
                Logger.WriteLineWarn($"EducationService CancelLiveCourseAsync Approved, roomId:{roomId}");
                await _educationDBService.UpdateCourseStatusEnumAsync(roomId, CourseStatusEnum.Approved);
                await _rtcService.CancelAsync(new CancelRequest
                {
                    LiveRoomCode = roomId,
                    UserCode = userCode,
                });
                return new LiveCourseBaseResult() { CourseCode = roomId };
            }
        }

        /// <summary>
        /// 学生获取观看地址
        /// </summary>
        /// <param name="request">观看地址请求实体</param>
        /// <errorCodes></errorCodes>
        /// <returns></returns>
        /// <docFirstDirectory>客户端API</docFirstDirectory>
        /// <docSecondDirectory>杏聆荟学院服务接口</docSecondDirectory>   
        /// <docThirdDirectory>学生视角</docThirdDirectory>  
        /// <errorCodes>LiveTeacherNotFound,CourseEnded,LiveCourseNotFound,CourseNotInProcess,StudentNotPay</errorCodes>
        public async Task<StudentLivePullUrlResult> GetCoursePullUrlAsync(InitiateLiveCourseRequest request)
        {
            var result = new StudentLivePullUrlResult();
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            if (string.IsNullOrEmpty(roomId))
            {
                ThrowRpcException(CustomerRpcCode.CourseCodeEmpty, "Course code empty");
            };
            using (await _lockManager.Acquire(roomId))
            {
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                if (courseDetailInfo == null)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                var courseName = courseDetailInfo.Name;
                if (courseDetailInfo.Status == CourseStatusEnum.HasEnded || courseDetailInfo.Status == CourseStatusEnum.Offline)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] course ended, current status:{courseDetailInfo.Status}, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseEnded, "Course ended");
                }

                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    ThrowRpcException(CustomerRpcCode.LiveCourseNotFound, "Live course not found");
                }
                var initiatorInfo = room.UserInfos.FirstOrDefault(c => c.Code == courseDetailInfo.TeacherCode) ?? new LiveMemberDTO();
                if (courseDetailInfo.Status != CourseStatusEnum.Started || initiatorInfo.Status != LiveMemberStatus.Joined)
                {
                    ThrowRpcException(CustomerRpcCode.CourseNotInProcess, "Course not in process");
                }

                //是否是付费课程
                var isNeedPay = courseDetailInfo.Price > 0 ? true : false;
                //验证是否收费
                if (courseDetailInfo.Experts != null && courseDetailInfo.Experts.Count > 0 && courseDetailInfo.Experts.Select(x => x.Code).Contains(userCode))
                {
                    courseDetailInfo.IsParticipant = true;
                }
                else if (courseDetailInfo.Assistants != null && courseDetailInfo.Assistants.Count > 0 && courseDetailInfo.Assistants.Select(x => x.Code).Contains(userCode))
                {
                    courseDetailInfo.IsAssistant = true;
                }
                if (!courseDetailInfo.IsParticipant && !courseDetailInfo.IsAssistant)
                {
                    var studentInfos = await _courseDBService.FindSignInListAsync(new List<string> { courseDetailInfo.Code });
                    var studentInfo = studentInfos.FirstOrDefault(c => c.StudentCode == userCode);
                    if (studentInfo == null)
                    {
                        ThrowRpcException(CustomerRpcCode.StudentNotPay, "Student not pay");
                    }
                    if (isNeedPay && !studentInfo.IsPay)
                    {
                        ThrowRpcException(CustomerRpcCode.StudentNotPay, "Student not pay");
                    }
                }
                result.RtmpPullUrl = courseDetailInfo.CourseChannel?.RtmpPullUrl;
                result.HlsPullUrl = courseDetailInfo.CourseChannel?.HlsPullUrl;
                result.HttpPullUrl = courseDetailInfo.CourseChannel?.HttpPullUrl;
            }
            return result;
        }

        /// <summary>
        /// 互动专家加入房间
        /// </summary>
        /// <param name="request">加入课程请求实体</param>
        /// <returns></returns>
        public async Task<JoinLiveCourseResult> JoinLiveCourseAsync(JoinLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            if (string.IsNullOrEmpty(roomId))
            {
                ThrowRpcException(CustomerRpcCode.CourseCodeEmpty, "Course code empty");
            };
            using (await _lockManager.Acquire(roomId))
            {
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                if (courseDetailInfo == null)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                var courseName = courseDetailInfo.Name;
                if (courseDetailInfo.Status == CourseStatusEnum.HasEnded || courseDetailInfo.Status == CourseStatusEnum.Offline)
                {
                    Logger.WriteLineWarn($"WingEducationService [{courseName}] course ended, current status:{courseDetailInfo.Status}, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.CourseEnded, "Course ended");
                }

                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    ThrowRpcException(CustomerRpcCode.LiveCourseNotFound, "Live course not found");
                }
                var initiatorInfo = room.UserInfos.FirstOrDefault(c => c.Code == courseDetailInfo.TeacherCode) ?? new LiveMemberDTO();
                if (courseDetailInfo.Status != CourseStatusEnum.Started)
                {
                    ThrowRpcException(CustomerRpcCode.CourseNotInProcess, "Course not in process");
                }
                var userCodes = new List<string>() { courseDetailInfo.TeacherCode };
                if (courseDetailInfo.Experts?.Count > 0)
                {
                    var expertUsers = courseDetailInfo.Experts.Select(c => c.Code).Distinct().ToList();
                    userCodes.AddRange(expertUsers);
                }
                if (courseDetailInfo.Assistants?.Count > 0)
                {
                    var assistantsUsers = courseDetailInfo.Assistants.Select(c => c.Code).Distinct().ToList();
                    userCodes.AddRange(assistantsUsers);
                }
                var memberRequest = new GetLiveMemberStatusInfosRequest()
                {
                    UserCodes = userCodes
                };
                var stautsInfo = await _rtcService.GetLiveMemberStatusInfosAsync(memberRequest);
                var curUserInfo = new LiveMemberStatusDTO();
                if (stautsInfo?.Count > 0)
                {
                    curUserInfo = stautsInfo.FirstOrDefault(c => c.Code == userCode) ?? new LiveMemberStatusDTO();
                }
                if (curUserInfo == null || (curUserInfo.UserStatusType == UserStatusEnum.IsBusy && curUserInfo.LiveRoomCode != roomId))
                {
                    Logger.WriteLineWarn($"WingEducationService userCode:[{userCode}] Join in live course failed, roomId:{roomId}");
                    ThrowRpcException(CustomerRpcCode.ApplyerIsBusy, "Applyer is busy");
                }
                //加入直播
                var acceptReq = new AcceptRequest()
                {
                    LiveRoomCode = roomId,
                    UserCode = userCode
                };
                var liveMember = await _rtcService.AcceptAsync(acceptReq);
                //获取签名
                var userSignRequest = new GetUserSignRequest
                {
                    UserId = userCode
                };
                var userSignResult = await _wingRtcService.GetUserSignAsync(userSignRequest);
                if (room.DeviceInfos?.Count > 0)
                {
                    foreach (var deviceInfo in room.DeviceInfos)
                    {
                        memberRequest.UserCodes.Add(deviceInfo.Code);
                    }
                }
                var statusList = await _rtcService.GetLiveMemberStatusInfosAsync(memberRequest);
                var memberStautsInfo = statusList?.Where(c => c.LiveRoomCode == roomId)?.ToList() ?? new List<LiveMemberStatusDTO>();
                ShareInfoDTO shareInfo = null;
                if (room.ShareInfo?.CurrentShareState == CourseShareStates.Opened)//正在共享
                {
                    shareInfo = room.ShareInfo;
                    //增加共享信息
                    //RTC信息到超声机的TRTC房间信息，以及输出宽高的信息
                }
                var result = new JoinLiveCourseResult
                {
                    CourseCode = roomId,
                    UserCode = userCode,
                    UserSign = userSignResult.UserSign,
                    AppId = room.SdkAppId,
                    RoomNo = room.RtcRoomId,
                    ShareInfo = room.ShareInfo,
                    HeartRateInterval = EnvironmentConfigManager.GetParammeter<IntParameter>("LiveConsultation", "HeartRateSeconds").Value,
                    LiveProtocol = EnvironmentConfigs.General.LiveProtocolDetail<TransactionStatusEnum>(),
                    MemberLiveDatas = TransferLiveMemberToCourseMember(room.UserInfos, courseDetailInfo.TeacherCode, memberStautsInfo, courseDetailInfo.Experts, courseDetailInfo.Assistants),
                    DeviceLiveDatas = TransferDeviceoCourseMember(room.DeviceInfos, memberStautsInfo),
                    CourseChannel = courseDetailInfo.CourseChannel,
                    OthersScreenSharingChannel = courseDetailInfo.OthersScreenSharingChannel,
                    RtcScreenSharingChannel = courseDetailInfo.RtcScreenSharingChannel,
                };
                return result;
            }
        }

        /// <summary>
        /// Send HeartRate request in a diagnosis metting
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        public async Task<LiveHeartRateResult> HeartRateAsync(LiveHeartRateRequest request)
        {
            var userToken = await _authenticationService.GetTokenAsync(request);
            var roomRequest = new GetLiveRoomRequest()
            {
                LiveRoomCode = request.LiveCode
            };
            var room = await _rtcService.GetLiveRoomAsync(roomRequest);
            if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
            {
                ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
            }
            var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = request.LiveCode });
            if (courseDetailInfo == null)
            {
                ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
            }
            if (courseDetailInfo.Status == CourseStatusEnum.NoApproval || courseDetailInfo.Status == CourseStatusEnum.RejectApproval)
            {
                ThrowRpcException(CustomerRpcCode.ReservationNotReviewed, $"Reservation record not reviewed, current status:{courseDetailInfo.Status.ToString()}");
            }
            if (courseDetailInfo.Status != CourseStatusEnum.Started)
            {
                ThrowRpcException(CustomerRpcCode.CourseEnded, $"Live Course Ended, current status:{courseDetailInfo.Status.ToString()}");
            }
            _liveHeartRateManager.AddOrUpdate(userToken.Code, userToken.ClientId, userToken.AccountName, request.LiveCode);
            return await Task.FromResult<LiveHeartRateResult>(new LiveHeartRateResult { LiveCode = request.LiveCode });
        }

        /// <summary>
        /// 暂时离开直播
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        /// <errorCodes>CourseNoFound,LiveTeacherNotFound</errorCodes>
        public async Task<LiveCourseBaseResult> LeaveLiveCourseAsync(LeaveLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            using (await _lockManager.Acquire(roomId))
            {
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
                }
                var liveMember = room.UserInfos?.FirstOrDefault(x => x.Code == userCode);
                if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
                {
                    ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course Live member not found");
                }
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                if (courseDetailInfo == null)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                var isClosed = true;
                var userCodes = new List<string>() { courseDetailInfo.TeacherCode };
                if (courseDetailInfo.Experts?.Count > 0)
                {
                    var expertUsers = courseDetailInfo.Experts.Select(c => c.Code).Distinct().ToList();
                    userCodes.AddRange(expertUsers);
                }
                if (courseDetailInfo.Assistants?.Count > 0)
                {
                    var assistantsUsers = courseDetailInfo.Assistants.Select(c => c.Code).Distinct().ToList();
                    userCodes.AddRange(assistantsUsers);
                }
                if (userCodes.Contains(userCode))
                {
                    //移除当前用户。因为当前用户肯定还没离线，看看其他的用户是否离线
                    userCodes.Remove(userCode);
                }
                if (room.UserInfos?.Count > 0)
                {
                    foreach (var item in room.UserInfos)
                    {
                        if (userCodes.Contains(item.Code) && (item.Status == LiveMemberStatus.Joined || item.Status == LiveMemberStatus.Accepted))
                        {
                            isClosed = isClosed & false;
                        }
                    }
                }
                var req = new LeaveRequest()
                {
                    LiveRoomCode = roomId,
                    UserCode = userCode,
                    Closed = isClosed
                };
                var res = await _rtcService.LeaveAsync(req);
                if (!res)
                {
                    Logger.WriteLineWarn($"LiveService LeaveLiveCourseAsync Error, userCode:{userCode}, roomId: {roomId}");
                }
                else
                {
                    if (isClosed)
                    {
                        Logger.WriteLineWarn($"EducationService LeaveLiveCourseAsync Approved, roomId:{roomId}");
                        await _educationDBService.UpdateCourseStatusEnumAsync(roomId, CourseStatusEnum.Approved);
                    }
                    await BrowsedRecordNotOnlineByCourseCode(roomId);                    
                }
                var result = new LiveCourseBaseResult() { CourseCode = roomId };
                return result;
            }
        }

        /// <summary>
        /// 结束课程或离开设置游客不在线
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        private async Task<bool> BrowsedRecordNotOnlineByCourseCode(string courseCode)
        {
            Logger.WriteLineWarn($"CourseLiveService BrowsedRecordNotOnlineByCourseCode courseCode:{courseCode}");
            var res = false;
            var dbRequest = new QueryBrowsedRecordPageRequest()
            {
                RelevanceCode = courseCode,
                PageIndex = 1,
                PageSize = 1000,
            };
            var result = await _educationDBService.QueryCourseBrowsedRecordsByDBAsync(dbRequest);
            if (result?.PageData?.Count > 0)
            {
                foreach(var item in result.PageData)
                {
                    //存在, 且在线，则更新为不在线
                    if (!string.IsNullOrEmpty(item?.CourseBrowsedRecordCode) && item.IsOnline)
                    {
                        _visitorLiveHeartRateManager.Close(item.UserCode, item.UserCode, item.DisplayName, item.RelevanceCode);
                        var updateEntity = item.MappingTo<CourseBrowsedRecordRequest>();
                        updateEntity.IsOnline = false;
                        res = await _educationDBService.VisitorBrowsedLeaveByDBAsync(updateEntity);                    
                    }
                }
            }
            return res;
        }

        /// <summary>
        /// 结束课程直播
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        public async Task<LiveCourseBaseResult> CompletedLiveCourseAsync(LeaveLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            using (await _lockManager.Acquire(roomId))
            {
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
                }
                var liveMember = room.UserInfos?.FirstOrDefault(x => x.Code == userCode);
                if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
                {
                    ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course live member not found");
                }
                var isClosed = true;
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                if (courseDetailInfo == null)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                if (courseDetailInfo.TeacherCode != userCode)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                // var userCodes = new List<string>() { courseDetailInfo.TeacherCode };
                // if (courseDetailInfo.Experts?.Count > 0)
                // {
                //     var expertUsers = courseDetailInfo.Experts.Select(c => c.Code).Distinct().ToList();
                //     userCodes.AddRange(expertUsers);
                // }
                // if (courseDetailInfo.Assistants?.Count > 0)
                // {
                //     var assistantsUsers = courseDetailInfo.Assistants.Select(c => c.Code).Distinct().ToList();
                //     userCodes.AddRange(assistantsUsers);
                // }
                // if (userCodes.Contains(userCode))
                // {
                //     //移除当前用户。因为当前用户肯定还没离线，看看其他的用户是否离线
                //     userCodes.Remove(userCode);
                // }
                var res = true;
                if (room.UserInfos?.Count > 0)
                {
                    var req = new LeaveRequest()
                    {
                        LiveRoomCode = roomId,
                        UserCode = userCode,
                        Closed = isClosed
                    };
                    var temp = await _rtcService.LeaveAsync(req);
                    res = res & temp;
                }
                if (res)
                {
                    Logger.WriteLineWarn($"EducationService CompletedLiveCourseAsync HasEnded, roomId:{roomId}");
                    //修改预约记录状态
                    var courseStatusRequest = new UpdateCourseStatusDBRequest()
                    {
                        Code = roomId,
                        Status = CourseStatusEnum.HasEnded
                    };
                    await _courseDBService.UpdateCourseStatusAsync(courseStatusRequest);
                    await BrowsedRecordNotOnlineByCourseCode(roomId);
                }
                var result = new LiveCourseBaseResult() { CourseCode = roomId };
                return result;
            }
        }

        /// <summary>
        /// 设置开启或关闭静音状态
        /// </summary>
        /// <param name="request">设置开启或关闭静音状态的请求</param>
        /// <returns></returns>
        public async Task<LiveCourseBaseResult> SetMuteLiveCourseAsync(SetMuteLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            var roomRequest = new GetLiveRoomRequest()
            {
                LiveRoomCode = roomId
            };
            var room = await _rtcService.GetLiveRoomAsync(roomRequest);
            if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
            {
                ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
            }
            if (room.Status != LiveRoomStatus.Connected && room.Status != LiveRoomStatus.Initiating)
            {
                ThrowRpcException(CustomerRpcCode.CourseNotInProcess, "Course Not InProcess");
            }
            else
            {
                var liveMember = room.UserInfos?.FirstOrDefault(x => x.Code == userCode);
                if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
                {
                    ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course Live member not found");
                }
                var req = new ChangeMuteOpenedRequest()
                {
                    LiveRoomCode = roomId,
                    UserCode = userCode,
                    MuteOpened = request.IsMute
                };
                await _rtcService.ChangeMuteOpenedAsync(req);
            }
            var result = new LiveCourseBaseResult() { CourseCode = roomId };
            return result;
        }

        /// <summary>
        /// 开启或关闭视频
        /// </summary>
        /// <param name="request">开启或关闭视频请求</param>
        /// <returns></returns>       
        public async Task<LiveCourseBaseResult> SetVideoLiveCourseAsync(SetVideoLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            var roomRequest = new GetLiveRoomRequest()
            {
                LiveRoomCode = roomId
            };
            var room = await _rtcService.GetLiveRoomAsync(roomRequest);
            if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
            {
                ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
            }
            if (room.Status != LiveRoomStatus.Connected && room.Status != LiveRoomStatus.Initiating)
            {
                ThrowRpcException(CustomerRpcCode.CourseNotInProcess, "Course Not InProcess");
            }
            else
            {
                var liveMember = room.UserInfos.FirstOrDefault(x => x.Code == userCode);
                if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
                {
                    ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course Live member not found");
                }
                var req = new ChangeVideoOpenedRequest()
                {
                    LiveRoomCode = roomId,
                    UserCode = userCode,
                    VideoOpened = request.IsOpened
                };
                await _rtcService.ChangeVideoOpenedAsync(req);
            }
            var result = new LiveCourseBaseResult() { CourseCode = roomId };
            return result;
        }

        /// <summary>
        /// 设置设备调参
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> SetControlParameterLiveCourseAsync(ControlParameterInCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            var roomRequest = new GetLiveRoomRequest()
            {
                LiveRoomCode = roomId
            };
            var room = await _rtcService.GetLiveRoomAsync(roomRequest);
            if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
            {
                ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
            }
            var liveMember = room.UserInfos.FirstOrDefault(x => x.Code == userCode);
            if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
            {
                ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course Live member not found");
            }
            var user = await _userDBService.FindUserByCodeAsync(userCode);
            //检查是否用用户使用参数调节中
            var isControllingMember = room.UserInfos.FirstOrDefault(x => x.IsControllingParameter);
            if (isControllingMember != null && isControllingMember.Code != userCode)
            {
                ThrowRpcException(CustomerRpcCode.ControllingParameterByOtherUser, "Controlling parameter by other");
            }
            var isControllingParameter = true;
            if (request.ControlType == ControlDeviceParameterEnum.End)
            {
                isControllingParameter = false;
            }
            if (liveMember.IsControllingParameter != isControllingParameter)
            {
                var req = new ChangeControllingStateRequest()
                {
                    LiveRoomCode = roomId,
                    UserCode = liveMember.Code,
                    DeviceCode = "",
                    IsControllingParameter = isControllingParameter
                };
                //todo 设备调参方法
                await _rtcService.ChangeControllingStateAsync(req);
            }
            return true;
        }

        /// <summary>
        /// 设备课程中加入房间
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        public async Task<LiveCourseMember> DeviceJoinInLiveCourseAsync(JoinLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            var deviceCode = request.DeviceCode;
            //获取房间缓存
            var roomId = request.CourseCode;
            if (string.IsNullOrEmpty(roomId))
            {
                ThrowRpcException(CustomerRpcCode.CourseCodeEmpty, "Course code empty");
            };
            using (await _lockManager.Acquire(roomId))
            {
                if (string.IsNullOrEmpty(request.DeviceCode))
                {
                    ThrowRpcException(CustomerRpcCode.DeviceCodeIsEmpty, "Device code is empty");
                }
                var cacheConsultationRecordDTO = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });//从rtc中取 device信息返回给客户端
                var deviceInfo = await _deviceDBService.FindDeviceInfoByCodeAsync(request.DeviceCode);
                var room = await _rtcService.GetLiveRoomAsync(new GetLiveRoomRequest { LiveRoomCode = roomId });
                bool deviceJoinResult = false;
                //如果房间中都设备已经在推流，joined，就比较已经加入房间了，返回已有的信息
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
                }
                //设备正在推流, 当前房间
                var deviceRtcInfo = room.DeviceInfos?.FirstOrDefault(x => x.Code == request.DeviceCode) ?? new LiveMemberDTO();
                if (!string.IsNullOrEmpty(deviceRtcInfo.Code))
                {
                    deviceJoinResult = true;
                }
                //没在当前房间
                if (!deviceJoinResult)
                {
                    // 查询是否被占用
                    // 获取用户状态，空闲、繁忙、离线
                    var deviceStatus = UserStatusEnum.NotOnline;
                    var userCodes = new List<string> { request.DeviceCode };
                    var clientInfos = _rtcService.GetLiveMemberStatusInfosAsync(new GetLiveMemberStatusInfosRequest { UserCodes = userCodes }).Result;
                    var client = clientInfos?.FirstOrDefault();
                    if (client != null)
                    {
                        deviceStatus = client.UserStatusType;
                    }
                    if (deviceStatus == UserStatusEnum.NotOnline)
                    {
                        ThrowRpcException(CustomerRpcCode.DeviceNotOnline, "Device not online");
                    }
                    if (deviceStatus == UserStatusEnum.IsBusy)
                    {
                        ThrowRpcException(CustomerRpcCode.DeviceIsBusy, "Device is busy");
                    }
                    //新增或更新设备直播间
                    var deviceName = deviceInfo.DisplayName;
                    var deviceLiveInfo = new LiveMemberDTO
                    {
                        Code = request.DeviceCode,
                        Name = deviceName,
                        MemberType = LiveMemberEnum.Device,
                        HeadImageToken = deviceInfo.HeadPicUrl,
                        Status = LiveMemberStatus.Accepted,
                    };
                    if (room.DeviceInfos == null)
                    {
                        room.DeviceInfos = new List<LiveMemberDTO>();
                    }
                    room.DeviceInfos.Add(deviceLiveInfo);
                    await _rtcService.AddOrUpdateLiveRoomAsync(new AddOrUpdateLiveRoomRequest
                    {
                        LiveRoom = room,
                    });
                }
                if (deviceRtcInfo?.Status != LiveMemberStatus.Joined)
                {
                    var acceptReq = new AcceptRequest()
                    {
                        LiveRoomCode = roomId,
                        DeviceCode = request.DeviceCode,
                        UserCode = userCode//邀请人，记录下来，用于离开时，带设备一起离线
                    };
                    var liveMember = await _rtcService.AcceptAsync(acceptReq);
                    room = await _rtcService.GetLiveRoomAsync(new GetLiveRoomRequest { LiveRoomCode = roomId });
                }
                var deviceLiveRoom = room.DeviceInfos?.FirstOrDefault(c => c.Code == request.DeviceCode);
                if (deviceLiveRoom == null)
                {
                    ThrowRpcException(CustomerRpcCode.DeviceNotExist, "Device not exist");
                }
                // if (deviceLiveRoom.Status == LiveMemberStatus.Default || deviceLiveRoom.Status == LiveMemberStatus.Accepted || deviceLiveRoom.Status == LiveMemberStatus.Left)
                // {
                //     deviceLiveRoom = await _rtcService.DeviceJoinInAsync(new DeviceJoinInRequest
                //     {
                //         LiveRoomCode = roomId,
                //         DeviceCode = request.DeviceCode
                //     });
                // }
                var videoDeviceOutputInfos = new List<VideoDeviceOutputInfo>();
                if (deviceInfo.VideoDeviceInfos?.Any() == true)
                {
                    foreach (var video in deviceInfo.VideoDeviceInfos)
                    {
                        var deviceVideoId = video.VideoDeviceId == deviceCode ? video.VideoDeviceId : $"{deviceInfo.ShortCode}_{video.VideoDeviceId}";
                        if (deviceInfo.VideoDeviceInfos.Count == 1 && !deviceInfo.VideoDeviceInfos.Any(x => x.VideoDeviceSourceType == (int)VideoDeviceSourceTypeEnum.Desktop))
                        {
                            deviceVideoId = deviceCode;
                        }
                        videoDeviceOutputInfos.Add(new VideoDeviceOutputInfo
                        {
                            VideoDeviceId = deviceVideoId,
                            VideoDeviceSourceType = (VideoDeviceSourceTypeEnum)video.VideoDeviceSourceType,
                            OutputWidth = video.OutputWidth,
                            OutputHeight = video.OutputHeight,
                        });
                    }
                }
                var courseMember = new LiveCourseMember()
                {
                    LiveData = deviceLiveRoom.LiveData,
                    LiveProtocol = EnvironmentConfigs.General.LiveProtocolDetail<TransactionStatusEnum>(),
                    Status = deviceLiveRoom.Status,
                    VideoDeviceInfos = videoDeviceOutputInfos,
                    MergedChannel = deviceInfo.MergedChannel,
                    MergedVideoOutputWidth = deviceInfo.MergedVideoOutputWidth,
                    MergedVideoOutputHeight = deviceInfo.MergedVideoOutputHeight,
                    IsTeacher = false,
                    IsExpertUser = false,
                    IsAssistantUser = false,
                    Mute = deviceLiveRoom.MuteOpened,
                    HeadImageToken = deviceLiveRoom.HeadImageToken,
                    MemberType = deviceLiveRoom.MemberType,
                    Name = deviceLiveRoom.Name,
                    Id = deviceLiveRoom.Code,
                    BackgroundRole = deviceLiveRoom.BackgroundRole,
                    VideoOpend = deviceLiveRoom.VideoOpened,
                    IsControllingParameter = deviceLiveRoom.IsControllingParameter
                };
                //推流告知其他的用户
                return courseMember;
            }
        }

        /// <summary>
        /// 设备课程中离开房间
        /// </summary>
        /// <param name="request">The request</param>
        /// <returns></returns>
        public async Task<LiveCourseBaseResult> DeviceLeaveLiveCourseAsync(DeviceLeaveLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            //获取房间缓存
            var roomId = request.CourseCode;
            using (await _lockManager.Acquire(roomId))
            {
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
                {
                    ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
                }
                var liveMember = room.DeviceInfos?.FirstOrDefault(x => x.Code == request.DeviceCode);
                if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
                {
                    ThrowRpcException(CustomerRpcCode.DeviceCodeIsEmpty, "Device code is empty");
                }
                var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                if (courseDetailInfo == null)
                {
                    ThrowRpcException(CustomerRpcCode.LiveTeacherNotFound, "Live teacher not find");
                }
                var req = new LeaveRequest()
                {
                    LiveRoomCode = roomId,
                    DeviceCode = request.DeviceCode,
                    UserCode = userCode,
                    Closed = false
                };
                var res = await _rtcService.LeaveAsync(req);
                if (!res)
                {
                    Logger.WriteLineWarn($"LiveService DeviceLeaveLiveCourseAsync Error, deviceCode:{request.DeviceCode}, roomId: {roomId}");
                }
                var result = new LiveCourseBaseResult() { CourseCode = roomId };
                return result;
            }
        }

        /// <summary>
        /// 发送白板交互数据
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> SendCourseInteractiveBoardDataAsync(SendLiveInteractiveBoardDataRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);
            var roomId = request.RoomId;
            var room = _rtcService.GetLiveRoomAsync(new GetLiveRoomRequest { LiveRoomCode = roomId });
            if (room == null)
            {
                ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
            }
            await _rtcService.SendBoardDataAsync(new SendBoardDataRequest
            {
                LiveRoomCode = roomId,
                UserCode = userCode,
                BoardData = request.BoardData,
                IsClear = request.IsClear,
            });
            return true;
        }


        /// <summary>
        /// 设置分享
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> SetShareInLiveCourseAsync(SetShareInLiveCourseRequest request)
        {
            var userCode = await GetClientIdByTokenAsync(request.Token);//操作
            //获取房间缓存
            var roomId = request.CourseCode;
            var roomRequest = new GetLiveRoomRequest()
            {
                LiveRoomCode = roomId
            };
            var livMemberName = "";
            var room = await _rtcService.GetLiveRoomAsync(roomRequest);
            if (room == null || string.IsNullOrEmpty(room.LiveRoomCode))
            {
                ThrowRpcException(CustomerRpcCode.CourseNoFound, "Course not existed");
            }
            if (room.Status != LiveRoomStatus.Connected && room.Status != LiveRoomStatus.Initiating)
            {
                ThrowRpcException(CustomerRpcCode.CourseNotInProcess, "Course Not InProcess");
            }
            else
            {
                var userLiveMember = room.UserInfos.FirstOrDefault(x => x.Code == userCode);
                if (userLiveMember == null || string.IsNullOrEmpty(userLiveMember.Code))
                {
                    ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course Live member not found");
                }
                //是否独占共享
                if (room.ShareInfo?.CurrentShareState == CourseShareStates.Opened)//正在共享
                {
                    ThrowRpcException(CustomerRpcCode.OtherCourseSharing, "Other Sharing");
                }
                using (await _lockManager.Acquire(roomId))
                {
                    if (request.CurrentShareType == CourseShareType.US)
                    {
                        if (request.CurrentShareState == CourseShareStates.Opened)
                        {
                            
                            Logger.WriteLineWarn($"SetShareInLiveCourseAsync 11, Request: {JsonConvert.SerializeObject(request)}, room: {JsonConvert.SerializeObject(room)}");
                            //调用超声机直播加入画面
                            var cacheConsultationRecordDTO = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });//从rtc中取 device信息返回给客户端
                            var deviceInfo = await _deviceDBService.FindDeviceInfoByCodeAsync(request.CurrentShareMemberCode);
                            bool deviceJoinResult = false;
                            //设备正在推流, 当前房间
                            var deviceRtcInfo = room.DeviceInfos?.FirstOrDefault(x => x.Code == request.CurrentShareMemberCode) ?? new LiveMemberDTO();
                            if (!string.IsNullOrEmpty(deviceRtcInfo.Code))
                            {
                                deviceJoinResult = true;
                            }
                            
                            Logger.WriteLineWarn($"SetShareInLiveCourseAsync 22, deviceJoinResult: {deviceJoinResult}, room: {JsonConvert.SerializeObject(room)}");
                            //没在当前房间
                            if (!deviceJoinResult)
                            {
                                // 查询是否被占用
                                var liveRoom = await _rtcService.GetInitiatingRoomByDeviceCodeAsync(new GetInitiatingRoomByDeviceCodeRequest
                                {
                                    DeviceCode = request.CurrentShareMemberCode
                                });
                                if (liveRoom != null && liveRoom.LiveRoomCode != roomId)
                                {
                                    ThrowRpcException(CustomerRpcCode.DeviceOtherRoomLiving, "Device other room living");
                                }
                                // 获取用户状态，空闲、繁忙、离
                                var deviceStatus = UserStatusEnum.NotOnline;
                                var userCodes = new List<string> { request.CurrentShareMemberCode };
                                var clientInfos = _rtcService.GetLiveMemberStatusInfosAsync(new GetLiveMemberStatusInfosRequest { UserCodes = userCodes }).Result;
                                var client = clientInfos?.FirstOrDefault();
                                if (client != null)
                                {
                                    deviceStatus = client.UserStatusType;
                                }
                                if (deviceStatus == UserStatusEnum.NotOnline)
                                {
                                    ThrowRpcException(CustomerRpcCode.DeviceNotOnline, "Device not online");
                                }
                                if (deviceStatus == UserStatusEnum.IsBusy)
                                {
                                    ThrowRpcException(CustomerRpcCode.DeviceIsBusy, "Device is busy");
                                }

                                //新增或更新设备直播间
                                var deviceName = deviceInfo.DisplayName;
                                var deviceLiveInfo = new LiveMemberDTO
                                {
                                    Code = request.CurrentShareMemberCode,
                                    Name = deviceName,
                                    MemberType = LiveMemberEnum.Device,
                                    HeadImageToken = deviceInfo.HeadPicUrl,
                                    Status = LiveMemberStatus.Accepted,
                                };
                                if (room.DeviceInfos == null)
                                {
                                    room.DeviceInfos = new List<LiveMemberDTO>();
                                }
                                room.DeviceInfos.Add(deviceLiveInfo);
                                var newData = await _rtcService.AddOrUpdateLiveRoomAsync(new AddOrUpdateLiveRoomRequest
                                {
                                    LiveRoom = room,
                                });
                                Logger.WriteLineWarn($"SetShareInLiveCourseAsync 33, AddOrUpdateLiveRoomAsync: {newData}, room: {JsonConvert.SerializeObject(room)}");
                                livMemberName = deviceName;
                            }
                            var acceptReq = new AcceptRequest()
                            {
                                LiveRoomCode = roomId,
                                DeviceCode = request.CurrentShareMemberCode,
                                UserCode = userCode//邀请人，记录下来，用于离开时，带设备一起离开
                            };
                            var liveMember = await _rtcService.AcceptAsync(acceptReq);
                            Logger.WriteLineWarn($"SetShareInLiveCourseAsync 44, liveMember: {liveMember}, room: {JsonConvert.SerializeObject(room)}");
                        }
                        else
                        {
                            Logger.WriteLineWarn($"SetShareInLiveCourseAsync aa, Request: {JsonConvert.SerializeObject(request)}, room: {JsonConvert.SerializeObject(room)}");
                            if (room.DeviceInfos?.Count > 0)
                            {
                                foreach(var deviceMemberInfo in room.DeviceInfos)
                                {
                                    if (string.IsNullOrEmpty(deviceMemberInfo?.Code))
                                    {
                                        ThrowRpcException(CustomerRpcCode.DeviceNotFind, "Live device not found");
                                    }

                                    //超声机离开房间
                                    var leaveReq = new LeaveRequest()
                                    {
                                        LiveRoomCode = roomId,
                                        DeviceCode = deviceMemberInfo.Code,
                                        UserCode = userCode,
                                        Closed = false
                                    };
                                    var res = await _rtcService.LeaveAsync(leaveReq);
                                    
                                    Logger.WriteLineWarn($"SetShareInLiveCourseAsync bb, res: {res}, deviceMemberInfo: {JsonConvert.SerializeObject(deviceMemberInfo)}");
                                }
                            }
                        }
                    }
                    else if (request.CurrentShareType == CourseShareType.Window)
                    {
                        var liveMember = room.UserInfos.FirstOrDefault(x => x.Code == request.CurrentShareMemberCode);
                        if (liveMember == null || string.IsNullOrEmpty(liveMember.Code))
                        {
                            ThrowRpcException(CustomerRpcCode.CourseLivMemberNotFound, "Course live member not found");
                        }
                        livMemberName = liveMember.Name;
                    }
                    var req = new ChangedShareInfoRequest()
                    {
                        OperatorCode = userCode,
                        LiveRoomCode = roomId,
                        ShareInfo = new ShareInfoDTO()
                        {
                            CurrentShareState = request.CurrentShareState,
                            CurrentShareType = request.CurrentShareType,
                            CurrentShareMemberCode = request.CurrentShareMemberCode,
                            CurrentShareMemberName = livMemberName
                        }
                    };
                    await _wingRtcService.ChangedShareInfoAsync(req);
                }

            }
            return true;
        }


        #region private method  
        /// <summary>
        /// 心跳检测进入房间
        /// </summary>
        /// <param name="roomId"></param>
        /// <param name="clientId"></param>
        /// <param name="name"></param>
        /// <returns></returns> 
        private async void HandleClientJoined(string roomId, string clientId, string name)
        {
            try
            {
                Logger.WriteLineInfo($"LiveService HandleClientJoined start name:{name}, clientId:{clientId}, roomId:{roomId}");
                //获取房间缓存
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                var liveMember = room?.UserInfos.FirstOrDefault(x => x.Code == clientId);
                if (liveMember == null)
                {
                    liveMember = room?.DeviceInfos.FirstOrDefault(x => x.Code == clientId);
                }
                
                Logger.WriteLineInfo($"LiveService HandleClientJoined liveMember:{JsonConvert.SerializeObject(liveMember)}");
                if (room != null && !string.IsNullOrEmpty(room.LiveRoomCode) && liveMember != null)
                {
                    if (liveMember.Status == LiveMemberStatus.Default || liveMember.Status == LiveMemberStatus.Accepted || liveMember.Status == LiveMemberStatus.Left)
                    {
                        var req = new HeartRateJoinRequest()
                        {
                            LiveRoomCode = roomId,
                            UserCode = liveMember.Code
                        };
                        await _rtcService.HeartRateJoinedAsync(req);
                    }
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"LiveService HandleClientJoined err, name:{name}, clientId:{clientId}, roomId:{roomId}, ex:{ex}");
            }
        }

        /// <summary>
        /// 心跳检测网络异常
        /// </summary>
        /// <param name="roomId"></param>
        /// <param name="clientId"></param>
        /// <param name="name"></param>
        /// <returns></returns> 
        private async void HandleClientNetworkErr(string roomId, string clientId, string name)
        {
            try
            {
                //获取房间缓存
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                var liveMember = room?.UserInfos.FirstOrDefault(x => x.Code == clientId);
                if (liveMember == null)
                {
                    liveMember = room?.DeviceInfos.FirstOrDefault(x => x.Code == clientId);
                }
                if (room != null && !string.IsNullOrEmpty(room.LiveRoomCode) && liveMember != null)
                {
                    if (liveMember.Status == LiveMemberStatus.Joined)
                    {
                        var req = new NetworkErrorRequest()
                        {
                            LiveRoomCode = roomId,
                            UserCode = liveMember.Code
                        };
                        await _rtcService.NetworkErrorAsync(req);
                    }
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"LiveService ClientNetworkErrAsync err, name:{name}, clientId:{clientId}, roomId:{roomId}, ex:{ex}");
            }
        }

        /// <summary>
        /// 心跳检测离线
        /// </summary>
        /// <param name="roomId"></param>
        /// <param name="clientId"></param>
        /// <param name="name"></param>
        /// <returns></returns> 
        private async void HandleClientLeave(string roomId, string clientId, string name)
        {
            try
            {                
                Logger.WriteLineInfo($"LiveService HandleClientLeave name:{name}, clientId:{clientId}, roomId:{roomId}");
                //获取房间缓存
                var roomRequest = new GetLiveRoomRequest()
                {
                    LiveRoomCode = roomId
                };
                var room = await _rtcService.GetLiveRoomAsync(roomRequest);
                var liveMember = room?.UserInfos.FirstOrDefault(x => x.Code == clientId);
                if (liveMember == null)
                {
                    liveMember = room?.DeviceInfos.FirstOrDefault(x => x.Code == clientId);
                }
                
                Logger.WriteLineInfo($"LiveService HandleClientLeave liveMember:{ JsonConvert.SerializeObject(liveMember) }");
                if (room != null && !string.IsNullOrEmpty(room.LiveRoomCode) && liveMember != null)
                {
                    if (liveMember.Status == LiveMemberStatus.Joined || liveMember.Status == LiveMemberStatus.Accepted)
                    {
                        var req = new HeartRateLeavedRequest()
                        {
                            LiveRoomCode = roomId,
                            UserCode = liveMember.Code
                        };
                        var res = await _rtcService.HeartRateLeavedAsync(req);
                        if (res)
                        {
                            bool isClosed = true;
                            // 发起人只能是老师
                            var courseDetailInfo = await _courseDBService.FindCourseByCodeAsync(new FindCourseByCodeDBRequest() { Code = roomId });
                            if (courseDetailInfo != null)
                            {
                                var userCodes = new List<string>() { courseDetailInfo.TeacherCode };
                                if (courseDetailInfo.Experts?.Count > 0)
                                {
                                    var expertUsers = courseDetailInfo.Experts.Select(c => c.Code).Distinct().ToList();
                                    userCodes.AddRange(expertUsers);
                                }
                                if (courseDetailInfo.Assistants?.Count > 0)
                                {
                                    var assistantsUsers = courseDetailInfo.Assistants.Select(c => c.Code).Distinct().ToList();
                                    userCodes.AddRange(assistantsUsers);
                                }
                                // if (userCodes.Contains(clientId))
                                // {
                                //     //移除当前用户。因为当前用户肯定还没离线，看看其他的用户是否离线
                                //     userCodes.Remove(clientId);
                                // }
                                var memberRequest = new GetLiveMemberStatusInfosRequest()
                                {
                                    UserCodes = userCodes
                                };
                                var memberStautsInfo = await _rtcService.GetLiveMemberStatusInfosAsync(memberRequest);
                                if (memberStautsInfo?.Count > 0)
                                {
                                    var statusList = memberStautsInfo.Where(c => c.LiveRoomCode == roomId)?.ToList() ?? new List<LiveMemberStatusDTO>();
                                    foreach (var item in statusList)
                                    {
                                        if (item.Status == LiveMemberStatus.Joined || item.Status == LiveMemberStatus.Accepted)
                                        {
                                            isClosed = isClosed & false;
                                        }
                                    }
                                }
                            }
                            
                            Logger.WriteLineInfo($"LiveService HandleClientLeave name:{name}, clientId:{clientId}, roomId:{roomId}, isClosed:{isClosed}");
                            if (isClosed)
                            {
                                // 修改预约记录状态
                                await _educationDBService.UpdateCourseStatusEnumAsync(roomId, CourseStatusEnum.Approved);
                                var updateLiveRoomReq = new CloseRequest()
                                {
                                    LiveRoomCode = roomId,
                                    UserCode = liveMember.Code,
                                };
                                var closeResult = await _rtcService.CloseAsync(updateLiveRoomReq);
                                if (closeResult)
                                {
                                    Logger.WriteLineWarn($"LiveService HandleClientLeave CloseRoom, name:{name}, clientId:{clientId}, roomId:{roomId}");
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"LiveService HandleClientLeave err, name:{name}, clientId:{clientId}, roomId:{roomId}, ex:{ex}");
            }
        }

        /// <summary>
        /// room房间成员与课程成员状态枚举类型转换
        /// </summary>
        /// <param name="status"></param>
        /// <returns></returns>
        private WingInterfaceLibrary.LiveConsultation.LiveConsultationMemberStatus TransferMemberStatus(LiveMemberStatusEnum status)
        {
            var obj = Enum.ToObject(typeof(WingInterfaceLibrary.LiveConsultation.LiveConsultationMemberStatus), (int)status);
            var statusEnum = (WingInterfaceLibrary.LiveConsultation.LiveConsultationMemberStatus)obj;
            return statusEnum;
        }

        /// <summary>
        /// 学员加入房间
        /// </summary>
        /// <param name="request">加入课程请求实体</param>
        /// <returns></returns>
        private IList<LiveCourseMember> TransferLiveMemberToCourseMember(IList<LiveMemberDTO> members, string teacherCode, IList<LiveMemberStatusDTO> memberStatusList, IList<StudentInfoDTO> experts, IList<StudentInfoDTO> assistants)
        {
            IList<LiveCourseMember> memberList = new List<LiveCourseMember>();
            if (members?.Count > 0)
            {
                var initiator = members.FirstOrDefault(c => c.Code == teacherCode) ?? new LiveMemberDTO();
                foreach (var item in members)
                {
                    var memberStatusInfo = memberStatusList?.FirstOrDefault(c => c.Code == item.Code) ?? new LiveMemberStatusDTO();
                    var isExpertUser = (experts.Count(c => c.Code == item.Code) > 0) ? true : false;
                    var isAssistantUser = (assistants.Count(c => c.Code == item.Code) > 0) ? true : false;
                    var isInitiator = (item.Code == teacherCode) ? true : false;
                    var tempLiveData = isInitiator ? initiator.LiveData : null;
                    var courseMember = new LiveCourseMember()
                    {
                        LiveData = tempLiveData,
                        LoginSource = memberStatusInfo.LoginSource,
                        LoginServerUrl = memberStatusInfo.LoginServerUrl,
                        Status = item.Status,
                        VideoDeviceInfos = new List<VideoDeviceOutputInfo>(),
                        IsTeacher = isInitiator,
                        IsExpertUser = isExpertUser,
                        IsAssistantUser = isAssistantUser,
                        Mute = item.MuteOpened,
                        UserStatusType = memberStatusInfo.UserStatusType,
                        HeadImageToken = item.HeadImageToken,
                        MemberType = item.MemberType,
                        Name = item.Name,
                        Id = item.Code,
                        VideoOpend = item.VideoOpened,
                        BackgroundRole = item.BackgroundRole,
                        IsControllingParameter = item.IsControllingParameter
                    };
                    memberList.Add(courseMember);
                }
            }
            return memberList;
        }

        /// <summary>
        /// 成员信息转换
        /// </summary>
        /// <param name="request">加入课程请求实体</param>
        /// <returns></returns>
        private IList<LiveCourseMember> TransferDeviceoCourseMember(IList<LiveMemberDTO> members, IList<LiveMemberStatusDTO> memberStatusList)
        {
            IList<LiveCourseMember> memberList = new List<LiveCourseMember>();
            if (members?.Count > 0)
            {
                foreach (var item in members)
                {
                    var deviceInfo = _deviceDBService.FindDeviceInfoByCodeAsync(item.Code).Result;
                    var videoDeviceOutputInfos = new List<VideoDeviceOutputInfo>();
                    if (deviceInfo.VideoDeviceInfos?.Any() == true)
                    {
                        var deviceCode = deviceInfo.DeviceCode;
                        foreach (var video in deviceInfo.VideoDeviceInfos)
                        {
                            var deviceVideoId = video.VideoDeviceId == deviceCode ? video.VideoDeviceId : $"{deviceInfo.ShortCode}_{video.VideoDeviceId}";
                            if (deviceInfo.VideoDeviceInfos.Count == 1 && !deviceInfo.VideoDeviceInfos.Any(x => x.VideoDeviceSourceType == (int)VideoDeviceSourceTypeEnum.Desktop))
                            {
                                deviceVideoId = deviceCode;
                            }
                            videoDeviceOutputInfos.Add(new VideoDeviceOutputInfo
                            {
                                VideoDeviceId = deviceVideoId,
                                VideoDeviceSourceType = (VideoDeviceSourceTypeEnum)video.VideoDeviceSourceType,
                                OutputWidth = video.OutputWidth,
                                OutputHeight = video.OutputHeight,
                            });
                        }
                    }

                    var memberStatusInfo = memberStatusList?.FirstOrDefault(c => c.Code == item.Code) ?? new LiveMemberStatusDTO();
                    var courseMember = new LiveCourseMember()
                    {
                        LiveData = item.LiveData,
                        LoginSource = memberStatusInfo.LoginSource,
                        LoginServerUrl = memberStatusInfo.LoginServerUrl,
                        Status = item.Status,
                        VideoDeviceInfos = videoDeviceOutputInfos,
                        IsTeacher = false,
                        IsExpertUser = false,
                        IsAssistantUser = false,
                        Mute = item.MuteOpened,
                        HeadImageToken = item.HeadImageToken,
                        MemberType = item.MemberType,
                        Name = item.Name,
                        Id = item.Code,
                        UserStatusType = memberStatusInfo.UserStatusType,
                        VideoOpend = item.VideoOpened,
                        BackgroundRole = item.BackgroundRole,
                        IsControllingParameter = item.IsControllingParameter
                    };
                    memberList.Add(courseMember);
                }
            }
            return memberList;
        }

        private async Task<ScreenSharingChannelDTO> GenerateRtcInfoAsync(int rtcRoomId, int sdkAppId)
        {
            try
            {
                var resultData = new ScreenSharingChannelDTO
                {
                    RtcRoomId = rtcRoomId,
                    SdkAppId = sdkAppId,
                };
                var userCode = Guid.NewGuid().ToString("N");
                resultData.UserCode = userCode;
                var userSignRequest = new GetUserSignRequest
                {
                    UserId = userCode
                };
                var userSignResult = await _wingRtcService.GetUserSignAsync(userSignRequest);
                resultData.UserSign = userSignResult.UserSign;
                var rtcResult = await _wingRtcService.GenerateRoomUrlAsync(new GenerateRoomUrlRequest
                {
                    UniqueId = userCode,
                    RoomId = (uint)rtcRoomId,
                });
                resultData.LiveData = new LiveDataDTO
                {
                    RtmpPushUrl = rtcResult.RtmpPushUrl,
                    RtmpPullUrl = rtcResult.RtmpUrl,
                    HlsPullUrl = rtcResult.HlsUrl,
                    HttpPullUrl = rtcResult.FlvUrl,
                };
                return resultData;
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"EducationService GenerateRtcInfoAsync err, ex:{ex}");
                return new ScreenSharingChannelDTO();
            }
        }

        /// <summary>
        /// 监听课程状态
        /// </summary>
        private bool ListenCourseStatus()
        {
            var result = false;
            try
            {
                //延期和已审核
                //查询直播，课程已开始，实际开始时间<= 当前时间的前一天
                var statuList = new List<CourseStatusEnum>()
                {
                    CourseStatusEnum.Approved,
                    CourseStatusEnum.Overdue
                };
                var waitCompletedCourses = _courseDBService.FindLiveCoursesByStatusDBAsync(statuList).Result;
                if (waitCompletedCourses?.Count > 0)
                {
                    foreach(var item in waitCompletedCourses)
                    {
                        var logReq = new QueryBrowsedRecordRequest()
                        {
                            RelevanceCode = item.Code,
                            UserCode = item.TeacherCode
                        };
                        var logInfo = _educationDBService.GetCourseBrowsedRecordByRelevanceCodeDBAsync(logReq).Result;
                        if (!string.IsNullOrEmpty(logInfo?.CourseBrowsedRecordCode) && logInfo.BrowsedTimeRecords?.Count > 0)
                        {
                            var curDT = DateTime.UtcNow.AddDays(-1);
                            var firstActualStartTime = logInfo.BrowsedTimeRecords.Min(c => c.StartTime);
                            if (firstActualStartTime > DateTime.MinValue && firstActualStartTime <= curDT)
                            {                                
                                Logger.WriteLineWarn($"EducationService ListenCourseStatus BrowsedRecord HasEnded, firstActualStartTime:{firstActualStartTime}, curDT:{curDT}");
                                //修改课程状态
                                var courseStatusRequest = new UpdateCourseStatusDBRequest()
                                {
                                    Code = item.Code,
                                    Status = CourseStatusEnum.HasEnded
                                };
                                var r = _courseDBService.UpdateCourseStatusAsync(courseStatusRequest).Result;
                                
                            }
                        }
                    }
                }
                //查询直播，课程已开始，实际开始时间<= 当前时间的前一天
                var DBRequest = new FindCoursePagesDBRequest()
                {
                    CourseType = CourseTypeEnum.LiveCourse,
                    Status = CourseStatusEnum.Started,
                    PageIndex = 1,
                    PageSize = 10000
                };
                var dataList = _courseDBService.FindCoursePagesAsync(DBRequest).Result;
                if (dataList?.PageData?.Count > 0)
                {
                    var dt = DateTime.UtcNow.AddDays(-1);
                    var findCourseList = dataList.PageData.Where(c => c.ActualStartTime <= dt)?.ToList();
                    if (findCourseList?.Count > 0)
                    {
                        foreach (var courseItem in findCourseList)
                        {
                            var userCodes = new List<string>() { courseItem.TeacherCode };
                            if (courseItem.Experts?.Count > 0)
                            {
                                var expertUsers = courseItem.Experts.Select(c => c.Code).Distinct().ToList();
                                userCodes.AddRange(expertUsers);
                            }
                            if (courseItem.Assistants?.Count > 0)
                            {
                                var assistantsUsers = courseItem.Assistants.Select(c => c.Code).Distinct().ToList();
                                userCodes.AddRange(assistantsUsers);
                            }
                            var clientInfos = _rtcService.GetLiveMemberStatusInfosAsync(new GetLiveMemberStatusInfosRequest { UserCodes = userCodes }).Result;
                            var noPersonInRoom = true;
                            if (clientInfos?.Count > 0)
                            {
                                foreach (var item in clientInfos)
                                {
                                    if (item.LiveRoomCode == courseItem.Code && userCodes.Contains(item.Code) && (item.Status == LiveMemberStatus.Joined || item.Status == LiveMemberStatus.Accepted))
                                    {
                                        noPersonInRoom = false;
                                        break;
                                    }
                                }
                            }
                            if (noPersonInRoom)
                            {
                                var updateLiveRoomReq = new CloseRequest()
                                {
                                    LiveRoomCode = courseItem.Code,
                                    UserCode = courseItem.TeacherCode,
                                };
                                var closeResult = _rtcService.CloseAsync(updateLiveRoomReq).Result;
                                if (closeResult)
                                {
                                    Logger.WriteLineWarn($"EducationService ListenCourseStatus CloseRoom, clientId:{courseItem.TeacherCode}, roomId:{courseItem.Code}");
                                }
                                Logger.WriteLineWarn($"EducationService ListenCourseStatus CloseRoom HasEnded, clientId:{courseItem.TeacherCode}, roomId:{courseItem.Code}");
                                //修改课程状态
                                var courseStatusRequest = new UpdateCourseStatusDBRequest()
                                {
                                    Code = courseItem.Code,
                                    Status = CourseStatusEnum.HasEnded
                                };
                                var r = _courseDBService.UpdateCourseStatusAsync(courseStatusRequest).Result;
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"EducationService ListenCourseStatus异常：" + ex.Message);
                return false;
            }
            return result;
        }

        /// <summary>
        /// 计算定时任务
        /// </summary>
        private void CaculateScheduleSet()
        {
            try
            {
                _timerCallPy = new System.Timers.Timer(_intervalMinutes * 60 * 1000);
                _timerCallPy.Elapsed += new ElapsedEventHandler(ExcuteSchedule);
                _timerCallPy.Start();
            }
            catch (Exception ex)
            {
                Console.WriteLine("CaculateScheduleSet计算定时任务异常：" + ex.Message);
            }
        }

        /// <summary>
        /// 定时器委托
        /// </summary>
        private void ExcuteSchedule(object sender, ElapsedEventArgs e)
        {
            try
            {
                ListenCourseStatus();
            }
            catch (Exception ex)
            {
                Console.WriteLine("ExcuteSchedule执行定时器委托：" + ex.Message);
            }
        }
        #endregion 
    }
}